// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


/*
 *
 * Revision : v1.0
 * Date     : 01/2010 - F.Locci
 * Objet    : Initial version - From previous Silecs-2
 * Action   :
 */

/*
 * This library was developped to provide communication between
 * Linux/LynxOS active host (client) and a PLC passive server.
 * Based on S7 or MODBUS std. protocole, it defines generic
 * set and get functions to read and write data into/from PLC
 * memory.
 *
 */

/* ---------------------------------------------------------*/
/* INCLUDE FILES																						*/
/* ---------------------------------------------------------*/
#include <fcntl.h>
#include <signal.h>

#include <silecs-communication/protocol/core/silecs.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ---------------------------------------------------------*/
/* MACRO DEFINITION																				  */
/* ---------------------------------------------------------*/

/* This macro is used to trap the unexpected broken pipe and
   return an error instead of exit process.
*/
static __sighandler_t sigpipeHandler = (__sighandler_t)-1;
#define _DISABLE_SIGPIPE sigpipeHandler = signal(SIGPIPE, SIG_IGN)
#define _ENABLE_SIGPIPE signal(SIGPIPE, sigpipeHandler)

/*----------------------------------------------------------*/
/* Time funtion
 * IeRfcSetTime: Convert lynxos time_t format to PLC _DT format
 * (SIMATIC).
 *
 * Details:
 * PLC DATE_AND_TIME (or DT) format is coded on 8 byte in BCD
 * format:
 * 0      1      2      3      4      5      6       7
 * YY     MM     DD     HH     MN     SC     SC/100  SC/10e4
 */
void IeRfcSetTime(unsigned char *dt, time_t epoch)  // to PLC
{
  struct tm *dscst;
  dscst = localtime(&epoch);

  dt[7] = 0;   // not used at the moment
  dt[6] = 0;   // not used at the moment
  dt[5] = _tobcd(dscst->tm_sec);
  dt[4] = _tobcd(dscst->tm_min);
  dt[3] = _tobcd(dscst->tm_hour);
  dt[2] = _tobcd(dscst->tm_mday);
  dt[1] = _tobcd(dscst->tm_mon+1);
  dt[0] = _tobcd(dscst->tm_year-100);
}

/*----------------------------------------------------------*/
/* Time funtion
 * IeRfcGetTime: Convert PLC _DT format to time_t format (SIMATIC).
 *
 * Details:
 * PLC DATE_AND_TIME (or DT) format is coded on 8 byte in BCD
 * format:
 * 0      1      2      3      4      5      6       7
 * YY     MM     DD     HH     MN     SC     SC/100  SC/10e4
 */
double IeRfcGetTime(unsigned char *dt) // from PLC
{
	struct tm plcst;
	time_t plctm;
	double ms;
	int year;

  plcst.tm_sec  = _frombcd(dt[5]);
  plcst.tm_min  = _frombcd(dt[4]);
  plcst.tm_hour = _frombcd(dt[3]);
  plcst.tm_mday = _frombcd(dt[2]);
  plcst.tm_mon  = _frombcd(dt[1])-1;
  year = _frombcd(dt[0]);
  //look at Siemens DATA_AND_TIME type documentation
  plcst.tm_year = ((year >= 90) ? year : year+100);

  ms = ((double)_frombcd(dt[6])/100.)+((double)_frombcd(dt[7])/10000.);
  plcst.tm_isdst = -1; // daylight saving time unavailable
  plctm = mktime(&plcst);

  return((double)plctm + ms);
}

/*----------------------------------------------------------*/
/* This is a function used to connect an host on a port.
 * 'ip' is the ip address string
 * 'port' is the port number (102 for rfc1006 server)
 * 'dst' is the TSAP destination string (exp: "TCP-1")
 * 'src' is the TSAP source string (exp: "TCP-1")
 * 'ts' timeout in second (1s minimum)
 * return a socket descriptor or <0 on error
 * (see constant error)
 *
 * Details:
 * intermediate connect_nonb() function is used to perform a non-blockant connection.
 * intermediate rfcPing() function is used to check if PLC is ON
 */
int connect_nonb(int sockfd, struct sockaddr *saptr, socklen_t salen, int nsec)
{
	int				flags, n, error;
	socklen_t		len;
	fd_set			rset, wset;
	struct timeval	tval;

	flags = fcntl(sockfd, F_GETFL, 0);
	fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);

	error = 0;
  /*v1.1*/
  _DISABLE_SIGPIPE;
	n = connect(sockfd, (struct sockaddr *) saptr, salen);
  _ENABLE_SIGPIPE;
	if (n < 0){
		if (errno != EINPROGRESS)
			return(-1);
	}

	/* Do whatever we want while the connect is taking place. */

	if (n == 0)
		goto done;	/* connect completed immediately */

	FD_ZERO(&rset);
	FD_SET(sockfd, &rset);
	wset = rset;
	tval.tv_sec = nsec;
	tval.tv_usec = 0;

	if ( (n = select(sockfd+1, &rset, &wset, NULL,
					 nsec ? &tval : NULL)) == 0) {
		close(sockfd);		/* timeout */
		errno = ETIMEDOUT;
		return(-1);
	}

	if (FD_ISSET(sockfd, &rset) || FD_ISSET(sockfd, &wset)) {
		len = sizeof(error);
		if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
			return(-1);			/* Solaris pending error */
		}
	}

done:
	fcntl(sockfd, F_SETFL, flags);	/* restore file status flags */

	if (error) {
		/*close(sockfd); just in case */
		errno = error;
		return(-1);
	}

	return(0);
}

/*----------------------------------------------------------*/
int rfcPing(char *ip, long ts)
{
	int err, s, val = 1;
	struct protoent *pent;
	struct sockaddr_in rsock;

	/* Socket create/connect */
	memset ((char *)&rsock,0,sizeof(rsock));

	if ((s=socket(AF_INET,SOCK_STREAM,0)) == -1)
	{
	  #ifdef COMMENT
		printf("#Error - Can't create socket: %s\n", strerror(errno));
	  #endif
		return RFC_SYS_ERROR;
	}

	/* Set TCP_NODELAY option to force immediate TCP/IP acknowledgement */
	if ((pent=getprotobyname("TCP")) == NULL)
	{
	  #ifdef COMMENT
		printf("#Error - Can't configure socket: %s\n", strerror(errno));
	  #endif
		return RFC_SYS_ERROR;
	}
	if (setsockopt(s,pent->p_proto,TCP_NODELAY,&val,4) == -1)
	{
	  #ifdef COMMENT
		printf("#Error - Can't configure socket: %s\n", strerror(errno));
	  #endif
		return RFC_SYS_ERROR;
	}

	#ifdef COMMENT
	printf("#Comment - Socket created: %i\n", s);
	#endif

	rsock.sin_addr.s_addr=inet_addr(ip);
	rsock.sin_family=AF_INET;
	/*check any port to detect if the host is OFF*/
	rsock.sin_port=htons(102);

	err = 0;
	if (connect_nonb(s,(struct sockaddr *)(&rsock),sizeof(rsock), ts) == -1)
	{
	  /*if hostname is OFF, connect() fails on TIMEOUT*/
      if ((errno == ETIMEDOUT) || (errno == EHOSTDOWN) || (errno == EHOSTUNREACH)) {
	  	err = RFC_TIMEOUT_ERROR;
	  }
	}

  	close(s);
	return err;
}

/*..........................................................*/
int IeRfcPing (char *hostName, char *plcIP)
{
  struct in_addr addr;
  struct hostent *hp;
  char *ipstr = plcIP;

  if ((hostName == NULL) && (plcIP == NULL))
	  return(IE_PARAM_ERROR);

  // use hostName reference in priority else plcIP
  if (hostName)
	{
	  hp = gethostbyname(hostName);
		if (hp)
		{
		  addr.s_addr = *((unsigned long int *) hp->h_addr);
		  ipstr = inet_ntoa(addr);
		}
	}

  if (ipstr == NULL)
 	return(IE_HOST_UNKNOWN_ERROR);

  /*trying to connect PLC (1 second max)*/
  return (rfcPing(ipstr, 1));
}

/*----------------------------------------------------------*/
/* This function return the string which describes the
 * error given by 'error' parameter.
 * 'error': error from ie library or PLC server (in)
 * return string pointer
 *
 * Details:
 */
char *IeGetErrorMessage(int err)
{
  int i;

  static struct
  {
	int ie_code;
	char * ie_mes;
  } errorMap[] =
  {
	{IE_SYS_ERROR, "[RFC lib.] function system error"},
	{IE_PARAM_ERROR, "[IE/RFC lib.] bad parameter(s) value/format"},
	{IE_CONNECT_ERROR, "[RFC lib.] PLC connection failed"},
	{IE_DISCONNECT_ERROR, "[RFC lib.] PLC disconnection requested"},
	{IE_TIMEOUT_ERROR, "[RFC lib.] communication (connect/send/recv) timeout occurred"},
	{IE_EPIPE_ERROR, "[RFC lib.] PLC communication broken"},
	{IE_FRAME_ERROR, "[RFC lib.] PLC data frame error"},
	{IE_DEFER_CMD_BUFFER_FULL, "[IE lib.] deferred commands buffer overflow"},
	{IE_DEFER_REP_BUFFER_EMPTY, "[IE lib.] deferred replies empty"},
	{IE_BLOCK_SIZE_ERROR, "[IE lib.] block size exceed PLC memory allocation"},
	{IE_NB_MONITORING_ERROR, "[IE lib.] monitoring descriptor out of limit"},
	{IE_HOST_PING_ERROR, "[IE lib.] PLC does not reply to ping"},
	{IE_HOST_UNKNOWN_ERROR, "[IE lib.] PLC hostname unknown"},
	{IE_CONFIG_FILE_ERROR, "[IE lib.] pair PLC/FEC not found in the config. file"},
	{IE_EQUIPMENT_UNKNOWN_ERROR, "[PLC] equipment id (devclass+eqp) unknown"},
	{IE_MEMORY_OVERFLOW_ERROR, "[PLC] get data buffer overflow"},
	{IE_ACCESS_VIOLATION_ERROR, "[PLC] DB unknown or get/set access violation"},
	{IE_MD_MONITORING_ERROR, "[PLC] bad monitoring descriptor or already used"},
	{IE_ALLOC_MONITORING_ERROR, "[PLC] monitoring record allocation failed"},
	{IE_UNALLOC_MONITORING_ERROR, "[PLC] monitoring record unallocation failed"},
	{IE_SET_TIME_IOCTL_ERROR, "[PLC] PLC date and time setting failed"},
	{IE_KEEP_ALIVE_IOCTL_ERROR, "[PLC] PLC keep-alive transactions failed"},
  };

  static int errorMapSize = sizeof(errorMap)/sizeof(errorMap[0]);

  if (err == 0)
	return ("No error");

  for (i=0; i<errorMapSize; i++)
	if (errorMap[i].ie_code == err)
	  return (errorMap[i].ie_mes);

  return ("Unknown IE error");
}

/*----------------------------------------------------------*/
/* This function initializes the DSC-PLC communication.
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 * Must be executed at first.
 */
int IeInit (void)
{
  int err=0;

  err |= IeMdbInit();

  return (err);
}

/*----------------------------------------------------------*/
/* This function de-initializes the DSC-PLC communication.
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 * Must be executed at end.
 */
int IeExit (void)
{
  int err=0;

  err |= IeMdbExit();

  return (err);
}

#ifdef __cplusplus
}
#endif

