// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
 * Revision : v1.0
 * Date     : 08/2003 - F.Locci
 * Objet    : original version for PS7-SCHNEIDER PLC
 * Action   :
 *
 * Revision : v1.1
 * Date     : 05/2004 - F.Locci
 * Objet    : .remove when data size = 0
 *			  .set err on read/write error
 *            Look at v1.1 comments
 *
 * Revision : v1.2
 * Date     : 11/2004 - F.Locci
 * Objet    : .move mb_master_init() call to IeMdbInit() function
 *            to make modbus code reentrant.
 *            Look at v1.2 comments
 *
 * Revision : v1.3
 * Date     : 10/2004 - F.Locci
 * Objet    : .Allow absolute offset for read/write action
 *			  .Allow variable size for read/write action
 *            => use offset sign convention for absolute/relative
 *            => use application size instead of config.table one
 *            Look at v1.3 comments
 *
 * Revision : v1.4
 * Date     : 03/2006 - F.Locci
 * Objet    : .Extend the number of possible connection (7 --> 32)
 *			  in order to allow more than 7 connections with
 *            DIFFERENT PLC (SPS/EA project in particular).
 *            We keep 7 connection  for SIEMENS (F/W) protocole
 *            to respect material constraint (14 ISO-ON-TCP max.)
 *
 * Revision : v1.5
 * Date     : 01/2007 - F.Locci
 * Objet    : implement BLOCK mode
 * Action   : .Add IeMdbGetBlock() and IeMdbSetBlock() function
 *            .Upgrade mdbFindEqp() function
 *            Look at v1.5 comments
 *
 * Revision : v1.6
 * Date     : 03/2007 - F.Locci
 * Objet    : Protect Get data buffer overflow
 *            Look at v1.5 comments
 *
 * Revision : v1.7
 * Date     : 04/2007 - F.Locci
 * Objet    : Direct byte swapping into MdbSetTime function
 *            Look at v1.5 comments
 */

/*
 * This library was developped to provide communication between
 * Linux/LynxOS active host (client) and a PLC passive server.
 * Based on MODBUS std. protocole, it defines generic
 * set and get functions to read and write data into/from PLC
 * memory.
 */

/* ---------------------------------------------------------*/
/* INCLUDE FILES																						*/
/* ---------------------------------------------------------*/
#include <silecs-communication/protocol/core/ietype.h>
#include "mb_addr.h"
#include "mb_master.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ---------------------------------------------------------*/
/* LOCAL CONSTANT DEFINITIONS																*/
/* ---------------------------------------------------------*/

/* For Modbus protocol */
#define MODBUS_MAX_CONNECTION    64 /*v1.4*/
#define MODBUS_SENDRETRIES_DEF   1
#define MODBUS_TIMEOUT_DEF       2    /* in seconds */

/*Config.Table entries offset - SCHNEIDER memory implementation - word alignment*/
#define FEC_CMD_OFS	      0    	/* Command from FEC: first entry of cfg.table*/
#define PARAMS_CMD_OFS	  2/2   /* Params of FEC command 		*/
#define CURRENT_DATE_OFS  10/2  /* PLC current dat&time */
#define SYS_INFO_OFS	  18/2  /* HW/SW info variables */
#define DIAG_INFO_OFS     78/2  /* PLC SW/HW diagnostic */
#define NB_EQP_OFS  	  132/2  /*nb table row */
#define ADDR_TBL_OFS  	  134/2  /*table data (addr/size)*/

#define swap16(w) ((((unsigned short)w & 0x00ff)<<8)+(((unsigned short)w & 0xff00)>>8))

/* ---------------------------------------------------------*/
/* TYPE DEFINITIONS					  															*/
/* ---------------------------------------------------------*/
/* an entry in the network array */
typedef struct
{
	const char      *name;
	node_addr_t     addr;
	int             nd; /* the node descriptor... */
	struct timespec timeout;
	u32				send_retries;
} network_t;

/* ---------------------------------------------------------*/
/* GLOBAL DECLARATIONS																			*/
/* ---------------------------------------------------------*/

/* ---------------------------------------------------------*/
/* FUNCTIONS CODE																						*/
/* ---------------------------------------------------------*/

/*----------------------------------------------------------*/
/* Local functions                                          */
/* ---------------------------------------------------------*/

/* Function to load a struct timeval correctly
 * from a double.
 */
static inline struct timespec d_to_timespec(double time) {
  struct timespec tmp;

  tmp.tv_sec  = time;
  tmp.tv_nsec = 1e9*(time - tmp.tv_sec);
  return tmp;
}

/*----------------------------------------------------------*/
/* Glocal functions                                         */
/* ---------------------------------------------------------*/

/*----------------------------------------------------------*/
/* Time funtion
 * IeMdbSetTime: Convert lynxos time_t format to PLC SCHNEIDER
 * format.
 *
 * Details:
 * PLC SCHNEIER format is coded on 8 bytes BCD
 * format:
 *
 *  0      1      2      3      4      5     6     7
 *  SC     00     HH     MN     MM     DD    YY	   YY
 *
 *  byte 1 is used by gateway to synchronize time setting.
 */
void IeMdbSetTime(unsigned char *dt, time_t epoch)  // to PLC
{
  struct tm *dscst;
  dscst = localtime(&epoch);

  /*v1.7*/
  /*In fact, each 16bits word is swapped with SCHNEIDER PLC
    1      0      3      2      5      4     7     6
    SC     00     HH     MN     MM     DD    YY	   YY

	Swapping can do it directly during data transferring
  */

  dt[1] = _tobcd(dscst->tm_sec);
  dt[0] = 0x00; /*not used*/
  dt[3] = _tobcd(dscst->tm_hour);
  dt[2] = _tobcd(dscst->tm_min);
  dt[5] = _tobcd(dscst->tm_mon+1);
  dt[4] = _tobcd(dscst->tm_mday);
  dt[7] = _tobcd(2000/100);
  dt[6] = _tobcd(dscst->tm_year-100);
}

/*----------------------------------------------------------*/
/* Time funtion
 * IeMdbGetTime: Convert PLC SCHNEIDER format to time_t format.
 *
 * Details:
 * PLC SCHNEIER format is coded on 8 bytes BCD
 * format:
 *
 *  0      1        2      3      4      5     6     7
 *  SC     SC/100   HH     MN     MM     DD    YY	 YY
 *
 */
double IeMdbGetTime(unsigned char *dt) // from PLC
{
  struct tm plcst;
  time_t plctm;
  double ms;
  int year;

  /*In fact, each 16bits word is swapped with SCHNEIDER PLC
    1      0        3      2      5      4     7     6
    SC     SC/100   HH     MN     MM     DD    YY	 YY

	Swapping can do it directly during data transferring
  */

  plcst.tm_sec  = _frombcd(dt[1]);
  plcst.tm_hour = _frombcd(dt[3]);
  plcst.tm_min  = _frombcd(dt[2]);
  plcst.tm_mon  = _frombcd(dt[5])-1;
  plcst.tm_mday = _frombcd(dt[4]);
  year = _frombcd(dt[6]);
  //look at Schneider DATA_AND_TIME type documentation
  plcst.tm_year = ((year >= 90) ? year : year+100);
  /*plcst.tm_wday = no used in PLC*/
  /*plcst.tm_yday = no used in PLC*/

  ms = ((double)_frombcd(dt[0])/100.);
  plcst.tm_isdst = -1; // daylight saving time unavailable
  plctm = mktime(&plcst);

  return((double)plctm + ms);
}

/*----------------------------------------------------------*/
/* This function initializes the DSC-PLC communication.
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 * Must be executed at first.
 */
int IeMdbInit (void)
{
  /*v1.2*/
  if (mb_master_init(MODBUS_MAX_CONNECTION) < 0)
  {
  #ifdef DEBUG
    /* could not init network! */
  	printf("\n#Error: mb_master_init() failed");
  #endif
    return (IE_SYS_ERROR);
  }

  return 0;
}

/*----------------------------------------------------------*/
/* This function de-initializes the DSC-PLC communication.
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 * Must be executed at end.
 */
int IeMdbExit (void)
{
  mb_master_done();
  return 0;
}

/*----------------------------------------------------------*/
/* This function creates a synchronous OSI-ON-TCP socket
 * channel used for sending/receiving communication with PLC.
 * 'hostName': Name of the host (in)
 * 'plcIP': PLC ip address string (in)
 * return socket descriptor or <0 on error (see const. error)
 *
 * Details:
 * 'plcIP' parameter is necessary only when 'hostName' is NULL.
 * Only one synchronous channel per PLC at the same time
 * supported in that version.
 */
int IeMdbOpen (char *hostName, char *plcIP, int baseAddress)
{
  network_t net;
  char *ipstr = plcIP;
  _WORD dummy;

  // use hostName reference in priority else plcIP
  if (hostName) ipstr = hostName;

  if (ipstr == NULL)
  {
    printf("#Error: ipstr NULL\n");
	return(IE_PARAM_ERROR);
  }

  /* init network .........................................*/
  /*net.name				= not used here*/
  net.addr.naf 				= naf_tcp;
  net.addr.addr.tcp.host	= ipstr;
  net.addr.addr.tcp.service = "502";       //specific modbus
  net.addr.addr.tcp.close_on_silence = 1;  // TRUE
  net.timeout       		= d_to_timespec(MODBUS_TIMEOUT_DEF);
  net.send_retries  		= MODBUS_SENDRETRIES_DEF;

  /* connection ......................................*/
  if ( (net.nd = mb_master_connect(net.addr)) >= 0)
  {
	  /*Try to access PLC memory, just to confirm modbus connection (for example,
		read config. table address.
	  */
  	  if (read_bdata(
					1,					/*int slave,*/
					baseAddress/2,	    /*int start_addr (/2 is used to convert the byte alignment to 16bit alignment) */
					sizeof(dummy),		/*int byte count,*/
					(_BYTE *)&dummy,	/*uchar *dest,*/
					net.nd,				/*int ttyfd,*/
					net.send_retries,	/*int send_retries,*/
					&net.timeout		/*const struct timespec *response_timeout*/
				) >= 0)
	  {
    	/* node opened successfully! */
		#ifdef DEBUG
    	printf("Connecting to network %s: OK\n", net.addr.addr.tcp.host);
		#endif

		return (net.nd);
	  }
	  #ifdef DEBUG
	  else
	  {
    	/* could not read registers from slave! */
    	printf("#Error: could not establish connection to %s\n", net.addr.addr.tcp.host);
	  }
	  #endif

  	  mb_master_close(net.nd);
  }
  #ifdef DEBUG
  else
  {
    /* could not connect to slave! */
    printf("#Error: could not establish connection to %s\n", net.addr.addr.tcp.host);
  }
  #endif

  return(IE_CONNECT_ERROR);
}

/*----------------------------------------------------------*/
int IeMBwriteData(int cid, _WORD dataOfs, _WORD dataSize, _BYTE* dataBuffer)
{
  int err;
  network_t net;

  /* init network .........................................*/
  net.timeout       = d_to_timespec(MODBUS_TIMEOUT_DEF);
  net.send_retries  = MODBUS_SENDRETRIES_DEF;

  /*then SET&DATE cmd into command word*/
  if (write_bdata(1,				/*int slave,*/
				  dataOfs,			/*int start_addr (!16 bits alignment)*/
				  dataSize,	        /*int byte count */
				  dataBuffer,	    /*uchar *src */
				  cid,	    		/*int connection id */
				  net.send_retries,	/*int send_retries,*/
				  &net.timeout		/*const struct timespec *response_timeout*/
				  ) >= 0)
  {
	#ifdef DEBUG
	printf("IeMBwriteData has been done successfully\n");
	#endif
	err = 0; /*no problem*/
  }
  else
  {
	#ifdef DEBUG
	  printf("#Error on IeMBwriteData\n");
	#endif
	err = IE_EPIPE_ERROR;
  }

  return err;
}

/*----------------------------------------------------------*/
int IeMBreadData(int cid, _WORD dataOfs, _WORD dataSize, _BYTE* dataBuffer)
{
  int err;
  network_t net;

  /* init network .........................................*/
  net.timeout       = d_to_timespec(MODBUS_TIMEOUT_DEF);
  net.send_retries  = MODBUS_SENDRETRIES_DEF;

  /*unusefull to read empty data buffer*/
  if (dataSize > 0)
  {
	 err = read_bdata(1,				/*int slave,*/
					  dataOfs,			/*int start_addr (!16 bits alignment)*/
					  dataSize,	        /*int byte count */
					  dataBuffer,	    /*uchar *src */
					  cid,	    		/*int connection id */
				      net.send_retries,	/*int send_retries,*/
				      &net.timeout		/*const struct timespec *response_timeout*/
				      );
  }

  if ((dataSize == 0) || (err >= 0))
  {
	#ifdef DEBUG
	printf("IeMBreadData has been done successfully\n");
	#endif
	err = 0; /*no problem*/
  }
  else
  {
	#ifdef DEBUG
	  printf("#Error on IeMBreadData\n");
	#endif
	err = IE_EPIPE_ERROR;
  }

  return err;
}

/*----------------------------------------------------------*/
/* This function closes the synchronous OSI-ON-TCP socket
 * channel.
 * 'sk': socket descriptor (in) - MUST BE a synch. pipe
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 */
int IeMdbClose(int cid)
{
  int err = 0;

  err |= mb_master_close(cid);

  if (err != 0) err = IE_DISCONNECT_ERROR;
  return(err);
}

#ifdef __cplusplus
}
#endif

