// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
 * Revision : v1.0
 * Date     : 08/2003 - F.Locci
 * Objet    : original version for PS7-SCHNEIDER PLC
 * Action   : 
 *
 * Revision : v1.5
 * Date     : 01/2007 - F.Locci
 * Objet    : implement BLOCK mode
 * Action   : Add IeMdbGetBlock() and IeMdbSetBlock() function 
 *            Look at v1.5 comments
 *
 */

#ifndef _IEMDB_H_
#define _IEMDB_H_

/*
 * This library was developped to provide communication between 
 * Linux/LynxOS active host (client) and a PLC passive server. 
 * Based on MODBUS std. protocole, it defines generic
 * set and get functions to read and write data into/from PLC 
 * memory.
  */

/* ---------------------------------------------------------*/
/* LOCAL CONSTANT DEFINITIONS                               */
/* ---------------------------------------------------------*/

/* ---------------------------------------------------------*/
/* TYPE DEFINITIONS                                         */
/* ---------------------------------------------------------*/

/* ---------------------------------------------------------*/
/* PROTOTYPE DEFINITIONS                                    */
/* ---------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

/*----------------------------------------------------------*/
/* Time funtion
 * IeMdbSetTime: Convert time_t epoch date to PLC SCHNEIDER
 * format.
 *
 * Details:
 * PLC SCHNEIER format is coded on 8 bytes BCD
 * format:
 *
 *  0      1      2      3      4      5     6     7 
 *  SC     --     HH     MN     MM     DD    YY	   YY
 *
 *  byte 1 is used by gateway to synchronize time setting.
 */
void IeMdbSetTime(unsigned char *dt, time_t epoch);  // to PLC

/*----------------------------------------------------------*/
/* Time funtion
 * IeMdbGetTime: Convert PLC SCHNEIDER format to lynxOS 
 * time_t format.
 *
 * Details:
 * PLC SCHNEIER format is coded on 8 bytes BCD
 * format:
 *
 *  0      1        2      3      4      5     6     7 
 *  SC     SC/100   HH     MN     MM     DD    YY	 YY
 *
 */
double IeMdbGetTime(unsigned char *dt); // from PLC

/*----------------------------------------------------------*/
/* This function initializes the DSC-PLC communication.
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 * Must be executed at first.
 */
int IeMdbInit (void);

/*----------------------------------------------------------*/
/* This function initializes the DSC-PLC communication.
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 * Must be executed at end.
 */
int IeMdbExit (void);

/*----------------------------------------------------------*/
/* This function creates a synchronous OSI-ON-TCP socket 
 * channel used for sending/receiving communication with PLC.
 * 'hostName': Name of the host (in)
 * 'plcIP': PLC ip address string (in)
 * return socket descriptor or <0 on error (see const. error)
 *
 * Details:
 * 'plcIP' parameter is necessary only when 'hostName' is NULL.
 * Only one synchronous channel per PLC at the same time 
 * supported in that version.
 */
int IeMdbOpen (char *hostName, char *plcIP, int baseAddress);

/*----------------------------------------------------------*/
/* This function send a data segment to the PLC using the
 * MODBUS protocole.
 * 'cid': write channel id of the MB connection
 * 'dataOfs': target adress of the data segment within the PLC memory
 * 'dataSize': byte size of the data segment to be sent
 * 'dataBuffer': buffer of the data to be sent
 * return 0 or <0 on error (see constant error)
 */
int IeMBwriteData(int cid, _WORD dataOfs, _WORD dataSize, _BYTE* dataBuffer);

/*----------------------------------------------------------*/
/* This function read a data segment from the PLC using the
 * MODBUS protocole.
 * 'cid': read channel id of the MB connection
 * 'dataOfs': adress of the data segment within the PLC memory
 * 'dataSize': byte size of the data segment to be read
 * 'dataBuffer': buffer to store the data
 * return 0 or <0 on error (see constant error)
 */
int IeMBreadData(int cid, _WORD dataOfs, _WORD dataSize, _BYTE* dataBuffer);

/*----------------------------------------------------------*/
/* This function closes the synchronous OSI-ON-TCP socket 
 * channel.
 * 'cid1': connection id (in) - MUST BE a synch. pipe
 * return 0 or <0 on error (see constant error)
 *
 * Details:
 */
int IeMdbClose(int cid);

#ifdef __cplusplus
}
#endif

#endif /* _IEMDB_H_ */
