/* mb_master.c

   By P.Costigan email: phil@pcscada.com.au http://pcscada.com.au

   These library of functions are designed to enable a program send and
   receive data from a device that communicates using the Modbus protocol.

   Copyright (C) 2000 Philip Costigan  P.C. SCADA LINK PTY. LTD.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


   The functions included here have been derived from the
   Modicon Modbus Protocol Reference Guide
   which can be obtained from Schneider at www.schneiderautomation.com.

   This code has its origins with
   paul@pmcrae.freeserve.co.uk (http://www.pmcrae.freeserve.co.uk)
   who wrote a small program to read 100 registers from a modbus slave.

   I have used his code as a catalist to produce this more functional set
   of functions. Thanks paul.



   10/2001 - Mario de Sousa
		Slightly re-organized the code.
                Miscelaneous cleanups
                Removed layer1 functions (i.e. write and read to /dev/ttySx)
                Added support for transaction id

                TODO
                add retries at layer2 (i.e. in this file)
                check response frame for correctness
*
*
* Revision : v1.0
* Date     : 11/2004 - F.Locci
* Objet    : .remove static prefix into modbus_tcp_write() declaration 
*            to make code reentrant.
*           .move global recv_buf_ to local one into read/write_bdata()
*            Look at v1.0 comments
*
* Revision : v2.0
* Date     : 11/2006 - F.Locci
* Objet    : swap16 bits from SILECS client instead of application level
*            Look at v2.0 comments
*/

#include <fcntl.h>	/* File control definitions */
#include <stdio.h>	/* Standard input/output */
#include <string.h>
#include <stdlib.h>
#include <termio.h>	/* POSIX terminal control definitions */
#include <time.h>	/* Time structures for select() */
#include <unistd.h>	/* POSIX Symbolic Constants */
#include <errno.h>	/* Error definitions */

#include <netinet/in.h> /* required for htons() and ntohs() */
#include "mb_layer1.h"
#include "mb_master.h"
#include "mb_master_private.h"
/*v1.0*/
#include "mb_tcp_private.h"

#ifdef __cplusplus
extern "C" {
#endif

/* #define DEBUG */		/* uncomment to see the data sent and received */


#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE 0
#endif


/******************************************/
/******************************************/
/**                                      **/
/**         Global Variables...          **/
/**                                      **/
/******************************************/
/******************************************/

/*v1.0
static u8 *query_buffer_ = NULL;
*/
#ifndef __Lynx__ 
#if RECV_BUFFER_SIZE < 5 //TCP_HEADER_LENGTH
#error The receive buffer is smaller than the frame header length.
#endif
#endif
/**/

/******************************************/
/******************************************/
/**                                      **/
/**       Local Utility functions...     **/
/**                                      **/
/******************************************/
/******************************************/


/*
 * Function to determine next transaction id.
 *
 * We use a library wide transaction id, which means that we
 * use a new transaction id no matter what slave to which we will
 * be sending the request...
 */
static inline u16 next_transaction_id(void) {

  static u16 next_id = 0;

  return next_id++;
}


/*
 * Functions to convert u16 variables
 * between network and host byte order
 *
 * NOTE: Modbus uses MSByte first, just like
 *       tcp/ip, so we use the htons() and
 *       ntoh() functions to guarantee
 *       code portability.
 */

static inline u16 mb_hton(u16 h_value) {
/*  return h_value; */
  return htons(h_value);
}

static inline u16 mb_ntoh(u16 m_value) {
/*  return m_value; */
  return ntohs(m_value);
}



static inline u8 msb(u16 value) {
/*  return Most Significant Byte of value; */
  return (value >> 8) & 0xFF;
}

static inline u8 lsb(u16 value) {
/*  return Least Significant Byte of value; */
  return value & 0xFF;
}

#define u16_v(char_ptr)  (*((u16 *)(&(char_ptr))))


/***********************************************************************

	The following functions construct the required query into
	a modbus query packet.

***********************************************************************/

static inline int build_query_packet(u8  slave,
                                     u8  function,
                                     u16 start_addr,
                                     u16 count,
                                     u8 *packet)
{
        packet[ 0 ] = slave,
        packet[ 1 ] = function;
          /* NOTE:
           *  Modbus uses high level addressing starting off from 1, but
           *  this is sent as 0 on the wire!
           */
        u16_v(packet[2]) = mb_hton(start_addr - 1);
        u16_v(packet[4]) = mb_hton(count);

        return 6;
}


/************************************************************************

	read_bdata

    F.Locci: just read raw byte array. Take care of frame which exceeds 
    modbus frame std. length.

*************************************************************************/
/*v2.0*/
#define swap16(w) ((((unsigned short)w & 0x00ff)<<8)+(((unsigned short)w & 0xff00)>>8))

int read_bdata(int slave,    /* slave id */
             int start_addr, /* word pointer addr. */
             int count,      /* byte count to read */
             u8 *dest,       /* byte array */
             int ttyfd,      /* connection id */
             int send_retries, /* number of request retries */
             const struct timespec *response_timeout) /* timeout value */
{
	/*v1.0*/
    u8 data[RECV_BUFFER_SIZE];
	u8 packet[QUERY_BUFFER_SIZE];
	/**/
	int rl, response_length=0;
    int query_length;
    u16 send_transaction_id, recv_transaction_id;
	u8  *destp  = dest;
	int global_length, length;            /* word count for read     */

    /*Attention! ModBus travaille sur une zone memoire [1..n] tandis que l'espace memoire
	  du PLC est definie [0..n-1] => start_addr = start_addr +1
	  start_addr est exprime en word pour le PLC (alignement 16bits).
	*/
	start_addr+=1;


	/*'global_length' is a word counter*/
	global_length = (count/2) + (count%2);
		  
    while(count) {

		if (count > MAX_READ_BDATA)
		   /*'length' is a word counter*/
		   length = MAX_READ_BDATA/2;
		else
		  /*'length' is a word counter*/
		  length = (count/2) + (count%2);
		  
        /* We must also initialize the recv_transaction_id with the same value,
         * since some layer 1 protocols do not support transaction id's, so
         * simply return the recv_transaction_id variable without any changes...
         */
        send_transaction_id = recv_transaction_id = next_transaction_id();

        query_length = build_query_packet(slave, 0x03, /*function*/
                                          start_addr, length,
                                          packet);
        if (query_length < 0)
          return -1;

		if (modbus_write(ttyfd, packet, query_length, send_transaction_id) < 0 )
        	  return PORT_FAILURE;

		rl = modbus_read(&ttyfd, data/*v1.0*/, &recv_transaction_id,
                         packet, query_length,
                         response_timeout);

		if( rl <= 0 )  return PORT_FAILURE;

        /* first check whther we have correct transaction id */
        if (send_transaction_id != recv_transaction_id)
          return PORT_FAILURE;

        if( rl > 0 )  
		{ rl -= 3; /*F.Locci, bug: 2->3*/
		  if (count<MAX_READ_BDATA)
		    rl -= (count%2);
		}

        if( rl > 0 ) {
		   memcpy(destp, (u8 *)&(data[3]), rl);
		   destp += rl;
		   start_addr += length; /*F.Locci, bug if rl = 0 */
		}
		
		/* start_addr += length;  F.Locci, bug if rl = 0 */
		response_length += rl;
		count -= rl;
	}

	/*v2.0:
	  Independamment de son mapping memoire (Little-endian), le PLC SCHNEIDER
	  swap chaque mot de 16bits qu'il transmet. Cette ligne permet de redresser
	  les donnees pour transmettre l'image exacte du mapping memoire. Charge 
	  a l'application ensuite d'appliquer le swapping necessaire relatif a son 
	  "Endianness".
	*/
	{
	int j;
	u16 *wdestp = (u16 *)dest; /* for swapping 16bit  */
	for(j=0; j<global_length; j++){ wdestp[j] = swap16(wdestp[j]); }
	}
	/**/

    #ifdef DEBUG
    { int i;
	printf("read_bdata: ");
	for(i=0; i<response_length; printf("%02x ", dest[i++]));
	printf("\n");
	}
	#endif
	
	return response_length;
}

/*************************************************************************

	write_bdata

    F.Locci: just write raw byte array. Take care of frame which exceeds 
    modbus frame std. length.
	!!byte count MUST be even value - 16 bits alignment in that PLC

***************************************************************************/

int write_bdata(int slave,   /* slave id */
             int start_addr, /* word pointer addr.  */
             int count,      /* byte count to write - MUST be even value */
             u8 *data,       /* byte array */
             int ttyfd,      /* connection id */
             int send_retries, /* number of request retries */
             const struct timespec *response_timeout) /* timeout value */
{
	int byte_count;
	int i, rl, response_length=0;
    int query_length;
	/*v1.0*/
    u8 rdata[RECV_BUFFER_SIZE];
	u8 packet[QUERY_BUFFER_SIZE];
    u16 send_transaction_id, recv_transaction_id;
	int length;        /* word count */
	u8 *destp, *srcp = data;

    /*Attention! ModBus travaille sur une zone memoire [1..n] tandis que l'espace memoire
	  du PLC est definie [0..n-1] => start_addr = start_addr +1
	  start_addr est exprime en word pour le PLC (alignement 16bits).
	*/
	start_addr+=1;
	
    while(count) {

		if (count > MAX_WRITE_BDATA) 
		{
		   /*'length' is a word counter*/
		   length = MAX_WRITE_BDATA/2;
		}
		else
		{
		  /*'length' is a word counter*/
		  length = (count/2) + (count%2);
		}

        /* We must also initialize the recv_transaction_id with the same value,
         * since some layer 1 protocols do not support transaction id's, so
         * simply return the recv_transaction_id variable without any changes...
         */
        send_transaction_id = recv_transaction_id = next_transaction_id();

        query_length = build_query_packet(slave, 0x10 /* function */,
                                          start_addr, length,
                                          packet);
        if (query_length < 0)
          return -1;

		byte_count = (length * 2);
		packet[ query_length ] = byte_count;

        destp = &(packet[ ++query_length ]);
		memcpy(destp, srcp, byte_count-(count%2));
		
		/*v2.0:
		  Independamment de son mapping memoire (Little-endian), le PLC SCHNEIDER
		  swap chaque mot de 16bits qu'il transmet. Cette ligne permet de redresser
		  les donnees pour transmettre l'image exacte du mapping memoire. Charge 
		  a l'application ensuite d'appliquer le swapping necessaire relatif a son 
		  "Endianness".
		*/
		{
		u16 *wdestp = (u16 *)destp;
		for(i=0; i<length; i++){ wdestp[i] = swap16(wdestp[i]); }
		}
		/**/

		srcp += byte_count-(count%2);
		query_length += byte_count;

		if( modbus_write( ttyfd, packet, query_length, send_transaction_id) < 0 )
		  return PORT_FAILURE;

		rl = modbus_read(&ttyfd, rdata/*v1.0*/, &recv_transaction_id,
                         packet, query_length, response_timeout);

        if( rl <= 0 ) return PORT_FAILURE;

        /* first check whther we have correct transaction id */
        if (send_transaction_id != recv_transaction_id)
          return PORT_FAILURE;
		  
		start_addr += length;
		response_length += byte_count;
		count -= (byte_count - (count%2));
    }

    return response_length;
}

/************************************************************************

	initialise / shutdown the library

	These functions sets up/shut down the library state
        (allocate memory for buffers, initialise data strcutures, etc)

**************************************************************************/

int mb_master_init(int nd_count) {
        int extra_bytes;

#ifdef DEBUG
	fprintf( stderr, "mb_master_init()\n");
	fprintf( stderr, "creating %d nodes\n", nd_count);
#endif

            /* initialise layer 1 library */
        if (modbus_init(nd_count, DEF_OPTIMIZATION, &extra_bytes)
            < 0)
          goto error_exit_0;

        /*v1.0 initialise send buffer
        query_buffer_ = (u8 *)malloc(QUERY_BUFFER_SIZE + extra_bytes);
        if (query_buffer_ == NULL)
          goto error_exit_1;
		*/  

        return 0;

error_exit_0:
        return -1;
}


int mb_master_done(void) {
/*v1.0
        free(query_buffer_);
        query_buffer_ = NULL;
*/
        return modbus_done();
}

/************************************************************************

	open/close master connection

	This function opens/closes a connection to the remote slave.

**************************************************************************/

int mb_master_connect(node_addr_t node_addr) {
#ifdef DEBUG
	fprintf( stderr, "mb_master_connect()\n");
#endif
          /* call layer 1 library */
        return modbus_connect(node_addr);
}


int mb_master_close(int nd) {
#ifdef DEBUG
	fprintf( stderr, "mb_master_close(): nd = %d\n", nd);
#endif
          /* call layer 1 library */
        return modbus_close(nd);
}

#ifdef __cplusplus
}
#endif
